#include "astnode.h"

template<class T>
static void generate_all(ostream& out, const deque<T *> batch)
{
  for (auto elt : batch)
    elt->generate(out);
}

static void print_line_control(ostream& out, const ast_node * node,
                        leaf_node * first = nullptr, bool do_it = true)
{
  if (do_it)
  {
    out << endl << "#line " << node->get_lineno() 
                << " \"" << node->get_filename() << "\"\n";
    if (first != nullptr)
      first->strip_preceding_newlines();
  }
}

void leaf_node::generate(ostream& out) const
{
  out << whitespace << text;
}

static inline string testname (const string& basename) 
{
  return "test_" + basename;
}

void test_suite::generate(ostream& out) const
{
  out << "#include <cppunit/TestFixture.h>" << endl
        << "#include <cppunit/TextTestRunner.h>" << endl
        << "#include <cppunit/extensions/HelperMacros.h>" << endl
        << endl;

  out << "/* THIS FILE IS AUTOGENERATED.  If you edit this file, your\n"
         " * changes will not be recorded in the source file.  Instead,\n"
         " * edit the .u file and re-run the test generator.\n"
         " * !!! DO NOT EDIT THIS FILE !!!\n"
         " */" << endl
         << endl;

  if (not prologue.empty())
    print_line_control(out, prologue.front());
  generate_all(out, prologue);

  string suitename = "Unci_" + suite_ident->get_text() + "_test";
  out << endl << endl
      << "class " << suitename << " : public CppUnit::TestFixture" << endl
      << "{" << endl;

  out << "  CPPUNIT_TEST_SUITE( " << suitename << " );" << endl;
  for (auto test : tests)
  {
    out << "    CPPUNIT_TEST( "
                << testname(test->get_name()) << " );" << endl;
  }
  out << "  CPPUNIT_TEST_SUITE_END();" << endl
        << endl;


  if (fixture == nullptr)
    out << endl << "  public:" << endl;
  else
    fixture->generate(out);

  generate_all(out, tests);

  out << endl
        << "};" << endl
        << endl;

  out << "int main( int argc, char **argv)" << endl
        << "{" << endl
        << "  CppUnit::TextTestRunner runner;" << endl
        << "  runner.addTest( " << suitename << "::suite() );" << endl
        << "  runner.run();" << endl
        << "  return 0;" << endl
        << "}" << endl;
}

void fixture_node::generate(ostream& out) const
{
  out << "  private:";
  generate_all(out, decls);
  if (setup_init != nullptr)
  {
    out << endl;
    out << "  void setUp_init()";
    setup_init->generate(out);
  }
  out << endl
        << "  public:" << endl;
  if (setup != nullptr)
  {
    out << endl;
    out << "  void setUp()";
    setup->generate(out);
  }
  if (teardown != nullptr)
  {
    out << endl;
    out << "  void tearDown()";
    teardown->generate(out);
  }
}

void named_block_node::generate(ostream& out) const
{
  // header actually generated by fixture
  block->generate(out);
}

void test_node::generate(ostream& out) const
{
  out << endl;
  out << "  void " << testname(name) << "()";
  block->generate(out);
}

void empty_stmt_node::generate(ostream& out) const
{
  print_line_control(out, this, semicolon, requires_line_sync);
  semicolon->generate(out);
}

void simple_stmt_node::generate(ostream& out) const
{
  print_line_control(out, this, elements.front(), requires_line_sync);
  generate_all(out, elements);
}

void decl_stmt_node::generate(ostream& out) const
{
  print_line_control(out, this, typedecl.front(), requires_line_sync);
  generate_all(out, typedecl);
  ident->generate(out);
  generate_all(out, rest);
}

void control_stmt_node::generate(ostream& out) const
{
  print_line_control(out, this, keyword, requires_line_sync);
  keyword->generate(out);
  generate_all(out, condition);
  block->generate(out);
}

void check_stmt_node::generate(ostream& out) const
{
  //out << whitespace;
  expectation->generate(out, expression);
}

void true_expectation_node::generate(ostream& out,
                                const deque<leaf_node *> expr) const
{
  print_line_control(out, expr.front(), expr.front());
  out << "    CPPUNIT_ASSERT ";
  generate_all(out, expr);
  out << ";";
}

void false_expectation_node::generate(ostream& out,
                                const deque<leaf_node *> expr) const
{
  print_line_control(out, expr.front(), expr.front());
  out << "    CPPUNIT_ASSERT (not ";
  generate_all(out, expr);
  out << ");";
}

void throw_expectation_node::generate(ostream& out,
                                const deque<leaf_node *> expr) const
{
  print_line_control(out, expr.front(), expr.front());
  out << "    CPPUNIT_ASSERT_ASSERTION_FAIL ( CPPUNIT_ASSERT ( ";
  generate_all(out, expr);
  out << " ))";
}

void about_expectation_node::generate(ostream& out,
                                const deque<leaf_node *> expr) const
{
  print_line_control(out, expr.front(), value.front());
  //print_line_control(out, value.front(), value.front());
  out << "    CPPUNIT_ASSERT_DOUBLES_EQUAL (";
  generate_all(out, value);
  out << ",";
  //print_line_control(out, expr.front(), expr.front());
  out << "        ";
  generate_all(out, expr);
  out << ",";
  //print_line_control(out, tolerance.front(), tolerance.front());
  out << "        ";
  generate_all(out, tolerance);
  out << ");";
}

void array_expectation_node::generate(ostream& out,
                                const deque<leaf_node *> expr) const
{
  out << "/* array-based check/expects not yet implemented */";
  //generate_all(out, expr);
}

void trace_expectation_node::generate(ostream& out,
                                const deque<leaf_node *> expr) const
{
  out << "/* trace-based check/expects not yet implemented */";
  //generate_all(out, expr);
}


void compop_expectation_node::generate(ostream& out,
                                const deque<leaf_node *> expr) const
{
  if (comp->get_text() == "==")
  {
    print_line_control(out, expr.front(), value.front());
    //print_line_control(out, value.front(), value.front());
    out << "    CPPUNIT_ASSERT_EQUAL ( (decltype";
    generate_all(out, expr);
    out << ")"; 
    //casting to expr's type ok b/c they're same or value is a plain
    //old integer, probably a literal, and expr is integral.  This is 
    //a little gross.
    generate_all(out, value);
    out << ",";
    //print_line_control(out, expr.front(), expr.front());
    out << "        ";
    generate_all(out, expr);
    out << " );";
  }
  else
  {
    print_line_control(out, expr.front(), expr.front());
    //print_line_control(out, expr.front(), expr.front());
    out << "    CPPUNIT_ASSERT ( ";
    generate_all(out, expr);
    comp->generate(out);
    //print_line_control(out, value.front(), value.front());
    out << "        ";
    generate_all(out, value);
    out << " );";
  }
}


void block_node::generate(ostream& out) const
{
  open->generate(out);
  generate_all(out, elements);
  close->generate(out);
}
