#include "astnode.h"
#include <set>

static void sem_err (ast_node * node, string msg)
{
  fprintf (stderr, "unci: %s at %s:%d\n", msg.c_str(),
                node->get_filename().c_str(), node->get_lineno());
  exit(1);
}

template<class T>
bool validate_all (deque<T> batch)
{
  bool retval = true;
  for (auto elt : batch)
    retval = elt->validate_and_transform() and retval; //don't short circuit
  return retval;
}

bool leaf_node::validate_and_transform()
{
  return true;
}

bool test_suite::validate_and_transform()
{
  bool retval =
    validate_all(prologue)
    && suite_ident->validate_and_transform()
    && (fixture == nullptr || fixture->validate_and_transform())
    && validate_all(tests);

  std::set<string> names;
  for (auto test : tests)
  {
    if (names.count(test->get_name()) > 0)
    {
      sem_err (test, "Duplicate test name " + test->get_name());
      retval = false;
    }
    names.insert(test->get_name());
  }

  return retval;
}

bool fixture_node::validate_and_transform()
{
  bool retval = true;
  if (setup != nullptr)
    retval = setup->validate_and_transform() && retval;
  if (teardown != nullptr)
    retval = teardown->validate_and_transform() && retval;

  deque<stmt_node *> setup_asgns;
  for (unsigned int i = 0; i < decls.size(); ++i)
  {
    //cerr << "Validating..." << endl;
    auto stmt = decls[i];
    auto decl = stmt->as_decl();
    if (decl == nullptr || decl->get_rest()[0]->get_text() != "=")
    {
      //if (dynamic_cast<simple_stmt_node*>(stmt[0])->
      // TODO: improve err msg when user has forgotten "tests:" and "test"
      //sem_err(stmt, "Bad assignment in fixture, should be of form 'type name = value;'");
      sem_err(stmt, "Only elements in fixture should be assignments of the form 'type name = value;', found something else");
      retval = false;
    }
    else
    {
      auto asgn = decl->duplicate_assignment();
      decls[i] = decl;
      setup_asgns.push_back(asgn);
      delete stmt;
      asgn->line_sync();
      decl->line_sync();
    }
  }

  if (retval and not setup_asgns.empty())
  {
    auto block = new block_node (filename, lineno - 1, setup_asgns);
    setup_init = new named_block_node(filename, lineno - 1, block);

    if (setup == nullptr)
    {
      setup = setup_init;
      setup_init = nullptr;
    }
    else
    {
      deque<leaf_node *> fake_stmt_leaf;
      fake_stmt_leaf.push_back(new leaf_node (setup->get_filename(),
                setup->get_lineno(), OTHER, "setUp_init();", "\n    "));
      auto fake_stmt = new simple_stmt_node(setup->get_filename(),
                setup->get_lineno(), fake_stmt_leaf);
      setup->block->prepend(fake_stmt);
      setup->lineno -= 1;
    }
  }

  return retval;
}

bool named_block_node::validate_and_transform()
{
  //cerr << "Check named block\n";
  if (not block->elements.empty())
    block->elements.front()->line_sync();

  return block->validate_and_transform();
}

bool test_node::validate_and_transform()
{
  bool sync = false;
  for (auto stmt : block->elements)
  {
    if (not sync)
    {
      stmt->line_sync();
      sync = true;
    }
    if (stmt->is_check())
      sync = false;
  }
  return block->validate_and_transform();
}

bool stmt_node::validate_and_transform()
{
  return true;
}

stmt_node * stmt_node::validation_stmt() const
{
  return nullptr;
}

bool control_stmt_node::validate_and_transform()
{
  bool retval = keyword->validate_and_transform();
  retval = validate_all(condition) && retval;
  retval = block->validate_and_transform() && retval;

  return retval;
}

bool block_node::validate_and_transform()
{
  //cerr << "Check block\n";
  bool retval =  validate_all(elements);

  for (auto it = elements.end(); it != elements.begin();)
  {
    --it;
    const auto& elt = *it;
    auto stmt = elt->validation_stmt();
    if (stmt != nullptr)
      it = elements.insert(it, stmt);
  }

  return retval;
}

bool check_stmt_node::validate_and_transform()
{
  //cerr << "Check val\n";
  bool retval = validate_all(expression);
  retval = expectation->validate_and_transform() && retval;

  return retval;
}

stmt_node * check_stmt_node::validation_stmt() const
{
  return expectation->validation_stmt(expression);
}

bool expectation_node::validate_and_transform()
{
  return true;
}

stmt_node * expectation_node::validation_stmt (const deque<leaf_node *>& expression) const
{
  return nullptr;
}

bool compop_expectation_node::validate_and_transform()
{
  bool all_value_strlit = true;
  for (const auto& val : value)
    if (val->get_token() != STRLIT)
      all_value_strlit = false;

  if (comp->get_text() == "=="
      and all_value_strlit)
  {
    /* special case: CPPUNIT_ASSERT_EQUAL does not permit comparison
     * of string and char*, but test cases are often of the form
     *   check (stringval) expect == "litstringval";
     * so this rewrites those as if they were
     *   check (stringval) expect == string("litstringval");
     * and since you won't normally be actually wanting to
     * check if something is == to a char* literal, this should 
     * be safe. */
    /* Also works for
     *   check (stringval) expect == "concatted" "string" "val";
     *   */
    auto strlit = value.front();
    value.push_front(new leaf_node(strlit->get_filename(),
                        strlit->get_lineno(), '(', "(", ""));
    value.push_front(new leaf_node(strlit->get_filename(),
                        strlit->get_lineno(), IDENT, "string",
                        strlit->get_whitespace()));
    value.push_back(new leaf_node(strlit->get_filename(),
                        strlit->get_lineno(), ')', ")", ""));
    strlit->set_whitespace("");
  }
  return true;
}

stmt_node * compop_expectation_node::validation_stmt (const deque<leaf_node *>& expression) const
{
  if (comp->get_text() == "==") 
  {
    deque<leaf_node *> expr;
    auto first = expression.front();
    expr.push_back(new leaf_node (first->get_filename(),
          first->get_lineno(), OTHER,
          "static_assert(std::is_same<decltype",
          "\n"));
    for (auto elt : expression)
      expr.push_back(elt->duplicate());
    expr.push_back(new leaf_node (first->get_filename(),
          first->get_lineno(), OTHER,
          ", decltype(",
          ""));
    for (auto elt : value)
      expr.push_back(elt->duplicate());
    expr.push_back(new leaf_node (first->get_filename(),
          first->get_lineno(), OTHER,
          ")>::value or (std::is_same<decltype(",
          ""));
    for (auto elt : value)
      expr.push_back(elt->duplicate());
    expr.push_back(new leaf_node (first->get_filename(),
          first->get_lineno(), OTHER,
          "), decltype(4)>::value and std::is_integral<decltype",
          ""));
    for (auto elt : expression)
      expr.push_back(elt->duplicate());
    expr.push_back(new leaf_node (first->get_filename(),
          first->get_lineno(), OTHER,
          ">::value), \"An expression being checked must be of the same type as the expected value\");",
          ""));

    auto retval = new simple_stmt_node(first->get_filename(),
        first->get_lineno(), expr);
    retval->line_sync();
    return retval;
  }
  else
    return nullptr;  // XXX
}

