#include "astnode.h"

#include <iostream>
using namespace std;


// methods for code generation found in astnode-generate.cpp


template <typename T>
static void delete_all_in (T &d)
{
  for (typename T::iterator it = d.begin();
       it != d.end();
       ++it)
    delete *it;
}

template <typename T>
static void print_all_prefix (const T &d, const string &prefix,
					const UnciLexer *lexer = nullptr)
{
  for (typename T::const_iterator it = d.begin();
       it != d.end();
       ++it)
  {
    (*it)->print_with_prefix(prefix, lexer);
  }
}

void leaf_node::print_with_prefix(const string &prefix,
					const UnciLexer *lexer) const
{
  cout << prefix << "LEAF " << lexer->token_name(token)
                        << " <" << whitespace << text << ">"
                                << endl;
}

leaf_node * leaf_node::duplicate() const
{
  return new leaf_node(filename, lineno, token, text, whitespace);
}

void leaf_node::strip_preceding_newlines()
{
  auto n = whitespace.find_last_of('\n');
  if (n != string::npos)
    whitespace.erase(0, n+1);
}

test_suite::~test_suite()
{
  delete_all_in(prologue);
  delete suite_ident;
  delete fixture;
  delete_all_in(tests);
}

void test_suite::print_with_prefix(const string &prefix,
					const UnciLexer *lexer) const
{
  cout << prefix << "PROLOGUE:\n";
  print_all_prefix (prologue, prefix + "  ", lexer);
  cout << prefix << "TEST SUITE " << suite_ident->get_text()
                        << ":" << endl;
  if (fixture != nullptr)
  {
    cout << prefix << "FIXTURE:\n";
    fixture->print_with_prefix(prefix + "  ", lexer);
  }
  if (not tests.empty())
  {
    cout << prefix << "TESTS:\n";
    for (auto test : tests)
      test->print_with_prefix(prefix + "  ", lexer);
  }
}

fixture_node::~fixture_node()
{
  delete_all_in(decls);
  delete setup;
  delete teardown;
}

void fixture_node::print_with_prefix(const string &prefix,
					const UnciLexer *lexer) const
{
  cout << prefix << "FIXTURE DECLARATIONS:\n";
  print_all_prefix (decls, prefix + "  ", lexer);
  if (setup != nullptr)
  {
    cout << prefix + "  " << "SETUP:\n";
    setup->print_with_prefix(prefix + "    ", lexer);
  }
  if (teardown != nullptr)
  {
    cout << prefix + "  " << "TEARDOWN:\n";
    teardown->print_with_prefix(prefix + "    ", lexer);
  }
}

named_block_node::~named_block_node()
{
  delete block;
}

void named_block_node::print_with_prefix(const string &prefix,
					const UnciLexer *lexer) const
{
  cout << prefix << "NAMED BLOCK:\n";
  block->print_with_prefix(prefix, lexer);
}

test_node::~test_node()
{
  delete block;
}

void test_node::print_with_prefix(const string &prefix,
					const UnciLexer *lexer) const
{
  cout << prefix << "TEST " << name << ":\n";
  block->print_with_prefix(prefix, lexer);
}

void empty_stmt_node::print_with_prefix(const string &prefix,
					const UnciLexer *lexer) const
{
  cout << prefix << "EMPTY STMT\n";
}

decl_stmt_node::~decl_stmt_node()
{
  delete_all_in(typedecl);
  delete ident;
  delete_all_in(rest);
}

void decl_stmt_node::print_with_prefix(const string &prefix,
					const UnciLexer *lexer) const
{
  cout << prefix << "DECL STMT:\n";
  print_all_prefix (typedecl, prefix + "  ", lexer);
  ident->print_with_prefix(prefix + "  ", lexer);
  print_all_prefix (rest, prefix + "  ", lexer);
}

simple_stmt_node * decl_stmt_node::split_decl()
{
  // copy of ident and all of rest to new stmt;
  // this one retains typedecl, ident, and gets new semicolon
  
  deque<leaf_node *> asgn = move(rest);
  rest.clear();
  rest.push_back(asgn.back()->duplicate()); // semicolon
  asgn.push_front(ident->duplicate()); // the identifier itself
  asgn.front()->set_whitespace("\n    ");
 
  auto first = asgn.front();
  return new simple_stmt_node(first->get_filename(), first->get_lineno(), asgn);
}

simple_stmt_node * decl_stmt_node::duplicate_assignment()
{
  // if this is decl-and-init, create assignment stmt to match
  // if decl only, return nullptr
  
  if (rest.front()->get_text() != "=")
    return nullptr;

  deque<leaf_node *> asgn;
  asgn.push_back(ident->duplicate());
  for (auto elt : rest)
  {
    asgn.push_back(elt->duplicate());
  }
  asgn.front()->set_whitespace("\n    ");
 
  auto first = asgn.front();
  return new simple_stmt_node(first->get_filename(), first->get_lineno(), asgn);
}

simple_stmt_node::~simple_stmt_node()
{
  delete_all_in(elements);
}

void simple_stmt_node::print_with_prefix(const string &prefix,
					const UnciLexer *lexer) const
{
  cout << prefix << "STMT:\n";
  print_all_prefix (elements, prefix + "  ", lexer);
}

decl_stmt_node * simple_stmt_node::as_decl()
{
  //cerr << "Validating: ";
  //for (auto elt : elements)
  //{
    //elt->generate(cerr);
  //}
  //cerr << endl;
  int ident_index = -1;
  int asgn_index = -1;
  for (unsigned int i = 0; i < elements.size(); ++i)
  {
    if (elements[i]->get_token() == IDENT)
    {
      if (asgn_index == -1)
        ident_index = i;
    }
    else if (elements[i]->get_token() == OPPUNC
                and elements[i]->get_text() == "=")
    {
      if (asgn_index == -1)
        asgn_index = i;
      else
        return nullptr;
    }
  }
  //cerr << "Indices: " << ident_index << "  " << asgn_index << endl;
  if (ident_index == -1
      or (asgn_index != -1 and ident_index + 1 != asgn_index))
    return nullptr;

  /* So stmt is either of the form
   *   blah blah blah ident ;
   * or
   *   blah blah blah ident = blah blah blah ;
   */

  deque<leaf_node *> typedecl;
  deque<leaf_node *> rest;
  for (unsigned int i = 0; i < elements.size(); ++i)
  {
    //assert (ident_index != -1);
    if (i < (unsigned) ident_index)
      typedecl.push_back(elements[i]);
    else if (i > (unsigned) ident_index)
      rest.push_back(elements[i]);
  }
  auto ident = elements[ident_index];
  elements.clear(); //so when this node gets deleted, the pieces don't

  return new decl_stmt_node (filename, lineno,
                        typedecl, ident, rest);
}

check_stmt_node::~check_stmt_node()
{
  delete_all_in(expression);
  delete expectation;
}

void check_stmt_node::print_with_prefix(const string &prefix,
					const UnciLexer *lexer) const
{
  cout << prefix << "CHECK STMT:\n";
  print_all_prefix (expression, prefix + "  ", lexer);
  cout << prefix << "EXPECT:\n";
  expectation->print_with_prefix(prefix + "  ", lexer);
}

control_stmt_node::~control_stmt_node()
{
  delete keyword;
  delete_all_in(condition);
  delete block;
}

void control_stmt_node::print_with_prefix(const string &prefix,
					const UnciLexer *lexer) const
{
  cout << prefix << "CONTROL STMT " << keyword->get_text() << ":\n";
  print_all_prefix (condition, prefix + "  ", lexer);
  cout << prefix << "DO:\n";
  block->print_with_prefix(prefix + "  ", lexer);
}

block_node::block_node (const string &f, int l, const deque<stmt_node *>& e)
      : stmt_node(f, l),  elements(e)
{
  open = new leaf_node(f, l, '{', "{", "\n  ");
  close = new leaf_node(f, l, '}', "}", "\n  ");
}

block_node::block_node (const string &f, int l)
      : stmt_node(f, l),  elements()
{
  open = new leaf_node(f, l, '{', "{", "\n  ");
  close = new leaf_node(f, l, '}', "}", "\n  ");
}

block_node::~block_node()
{
  delete open;
  delete_all_in(elements);
  delete close;
}

void block_node::print_with_prefix(const string &prefix,
					const UnciLexer *lexer) const
{
  cout << prefix << "BLOCK:\n";
  open->print_with_prefix(prefix + "  ", lexer);
  print_all_prefix (elements, prefix + "  ", lexer);
  close->print_with_prefix(prefix + "  ", lexer);
}

//void block_node::prepend(const deque<stmt_node *> other)
//{
  //elements.insert(elements.begin(), other.begin(), other.end());
//}

void true_expectation_node::print_with_prefix(const string &prefix,
					const UnciLexer *lexer) const
{
  cout << prefix << "TRUE\n";
}

void false_expectation_node::print_with_prefix(const string &prefix,
					const UnciLexer *lexer) const
{
  cout << prefix << "FALSE\n";
}

void throw_expectation_node::print_with_prefix(const string &prefix,
					const UnciLexer *lexer) const
{
  cout << prefix << "THROW\n";
}

about_expectation_node::~about_expectation_node()
{
  delete_all_in(value);
  delete_all_in(tolerance);
}

void about_expectation_node::print_with_prefix(const string &prefix,
					const UnciLexer *lexer) const
{
  cout << prefix << "ABOUT\n";
  print_all_prefix (value, prefix + "  ", lexer);
  cout << prefix << "PLUS OR MINUS\n";
  print_all_prefix (tolerance, prefix + "  ", lexer);
}

array_expectation_node::~array_expectation_node()
{
  delete_all_in(value);
}

void array_expectation_node::print_with_prefix(const string &prefix,
					const UnciLexer *lexer) const
{
  cout << prefix << "ARRAY something\n";
}

void trace_expectation_node::print_with_prefix(const string &prefix,
					const UnciLexer *lexer) const
{
  cout << prefix << "TRACE\n";
}

compop_expectation_node::~compop_expectation_node()
{
  delete comp;
  delete_all_in(value);
}

void compop_expectation_node::print_with_prefix(const string &prefix,
					const UnciLexer *lexer) const
{
  cout << prefix << "IS " << comp->get_text() << " TO/THAN:" << endl;
  print_all_prefix (value, prefix + "  ", lexer);
}

