#ifndef _ASTNODE_H_
#define _ASTNODE_H_

#include <string>
#include <deque>
using namespace std;

#include "tokens.h"
#include "uncilexer.h"

class ast_node;
class leaf_node;
class test_suite;
class fixture_node;
class named_block_node;
class test_node;

class stmt_node;
class empty_stmt_node;
class decl_stmt_node;
class simple_stmt_node;
class check_stmt_node;
class control_stmt_node;
class block_node;

class expectation_node;
class true_expectation_node;
class false_expectation_node;
class throw_expectation_node;
class about_expectation_node;
class array_expectation_node;
class trace_expectation_node;
class compop_expectation_node;

class ast_node 
{
  public:
    ast_node(const string &f, int l)
      : filename(f), lineno(l) { }
    virtual ~ast_node() { }

    virtual void print_with_prefix (const string &prefix,
                                const UnciLexer *lexer = nullptr) const = 0;

    virtual bool validate_and_transform () = 0;
    virtual void generate (ostream& out) const = 0;

    inline string get_filename() const { return filename; }
    inline int get_lineno() const { return lineno; }

  protected:
    string filename;
    int lineno;
    string whitespace;
};


class leaf_node : public ast_node
{
  public:
    leaf_node(const string &f, int l, int tok, const string& t, const string& ws)
      : ast_node(f, l), token(tok), text(t) { whitespace=ws; }
    virtual ~leaf_node() { }

    virtual void print_with_prefix (const string &prefix,
				const UnciLexer *lexer = nullptr) const;
    virtual bool validate_and_transform ();
    virtual void generate (ostream& out) const;

    leaf_node * duplicate() const;

    token_t get_token() const { return token; }
    string get_text() const { return text; }
    string get_whitespace() const { return whitespace; }
    void set_whitespace(string s) { whitespace = s; }
    void strip_preceding_newlines();

  private:
    token_t token;
    string text;
    //string whitespace;  //i.e. preceding whitespace (now part of ast_node)
};


class test_suite : public ast_node
{
  public:
    test_suite (const string& filename, int lineno, deque<ast_node *> p,
                leaf_node * si, fixture_node * f, deque<test_node *> t)
      : ast_node (filename, lineno), prologue(p), suite_ident(si),
                        fixture(f), tests(t) { }
    virtual ~test_suite();

    virtual void print_with_prefix (const string &prefix,
				const UnciLexer *lexer = nullptr) const;
    virtual bool validate_and_transform ();
    virtual void generate (ostream& out) const;

  private:
    deque<ast_node *> prologue;   // func_def, var_decl
    leaf_node * suite_ident;
    fixture_node * fixture;
    deque<test_node *> tests;
};

class stmt_node : public ast_node
{
  public:
    stmt_node (const string &f, int l)
      : ast_node(f, l), requires_line_sync(false) { }
    virtual ~stmt_node() {}

    virtual void print_with_prefix (const string &prefix,
			const UnciLexer *lexer = nullptr) const = 0;
    virtual bool validate_and_transform ();
    virtual stmt_node * validation_stmt () const;

    virtual decl_stmt_node * as_decl() { return nullptr; }

    inline virtual bool is_check() const { return false; }

    void line_sync() { requires_line_sync = true; }

  protected:
    bool requires_line_sync;
};

class empty_stmt_node : public stmt_node
{
  public:
    empty_stmt_node (const string &f, int l, leaf_node * s)
      : stmt_node(f, l), semicolon(s) { }
    virtual ~empty_stmt_node() {}

    virtual void print_with_prefix (const string &prefix,
			        const UnciLexer *lexer = nullptr) const;
    virtual void generate (ostream& out) const;
  private:
    leaf_node * semicolon;
};

class decl_stmt_node : public stmt_node
{
  public:
    decl_stmt_node (const string &f, int l,
                deque<leaf_node *>& t, leaf_node * i,
                deque<leaf_node *>& r)
      : stmt_node(f, l), typedecl(t), ident(i), rest(r) { }
    virtual ~decl_stmt_node();

    virtual void print_with_prefix (const string &prefix,
        		        const UnciLexer *lexer = nullptr) const;
    virtual void generate (ostream& out) const;


    deque<leaf_node *> get_rest() const { return rest; }
    simple_stmt_node * split_decl();
    simple_stmt_node * duplicate_assignment();

  private:
    deque<leaf_node *> typedecl;
    leaf_node * ident;
    deque<leaf_node *> rest;
};

class simple_stmt_node : public stmt_node
{
  public:
    simple_stmt_node (const string &f, int l, const deque<leaf_node *>& e)
      : stmt_node(f, l), elements(e) { }
    virtual ~simple_stmt_node();

    virtual void print_with_prefix (const string &prefix,
        		        const UnciLexer *lexer = nullptr) const;
    virtual void generate (ostream& out) const;

    virtual decl_stmt_node * as_decl();

  private:
    deque<leaf_node *> elements;
};

class check_stmt_node : public stmt_node
{
  public:
    check_stmt_node (const string &f, int l, string ws,
                        const deque<leaf_node *>& e, expectation_node * ex)
      : stmt_node(f, l), expression(e), expectation(ex) { whitespace = ws; }
    virtual ~check_stmt_node();

    virtual void print_with_prefix (const string &prefix,
        		        const UnciLexer *lexer = nullptr) const;
    virtual bool validate_and_transform ();
    virtual stmt_node * validation_stmt () const;
    virtual void generate (ostream& out) const;

    inline virtual bool is_check() const { return true; }

  private:
    //string whitespace;  // now in ast_node
    deque<leaf_node *> expression;
    expectation_node * expectation;
};

class control_stmt_node : public stmt_node
{
  public:
    control_stmt_node (const string &f, int l, leaf_node * k,
                                const deque<leaf_node *>& c, stmt_node * b)
      : stmt_node(f, l), keyword(k), condition(c), block(b) { }
    virtual ~control_stmt_node();

    virtual void print_with_prefix (const string &prefix,
        		        const UnciLexer *lexer = nullptr) const;
    virtual bool validate_and_transform ();
    virtual void generate (ostream& out) const;

  private:
    leaf_node * keyword;
    deque<leaf_node *> condition;
    stmt_node * block;
};

class block_node : public stmt_node
{
  public:
    block_node (const string &f, int l,
                leaf_node * o, const deque<stmt_node *>& e, leaf_node * c)
      : stmt_node(f, l), open(o), elements(e), close(c) { }
    block_node (const string &f, int l);
    block_node (const string &f, int l, const deque<stmt_node *>& e);
    virtual ~block_node();

    virtual void print_with_prefix (const string &prefix,
				const UnciLexer *lexer = nullptr) const;
    virtual bool validate_and_transform ();
    virtual void generate (ostream& out) const;

    //void prepend(const deque<stmt_node *> other);
    void prepend(stmt_node * other) { elements.push_front(other); }

    friend class named_block_node;
    friend class test_node;

  private:
    leaf_node * open;
    deque<stmt_node *> elements;
    leaf_node * close;
};

class expectation_node : public ast_node
{
  public:
    expectation_node (const string &f, int l)
      : ast_node(f, l) { }
    virtual ~expectation_node() { }

    virtual void print_with_prefix (const string &prefix,
			const UnciLexer *lexer = nullptr) const = 0;
    virtual bool validate_and_transform ();
    virtual stmt_node * validation_stmt (const deque<leaf_node *>& expression) const;
    virtual void generate (ostream& out) const {
      throw ("Internal error (incorrect call to generate)");
    }
    virtual void generate (ostream& out,
          const deque<leaf_node *> expr) const = 0;

  private:
};

class true_expectation_node : public expectation_node
{
  public:
    true_expectation_node (const string &f, int l)
      : expectation_node(f, l) { }
    virtual ~true_expectation_node() { }

    virtual void print_with_prefix (const string &prefix,
				const UnciLexer *lexer = nullptr) const;
    virtual void generate (ostream& out,
                                const deque<leaf_node *> expr) const;

  private:
};

class false_expectation_node : public expectation_node
{
  public:
    false_expectation_node (const string &f, int l)
      : expectation_node(f, l) { }
    virtual ~false_expectation_node() { }

    virtual void print_with_prefix (const string &prefix,
				const UnciLexer *lexer = nullptr) const;
    virtual void generate (ostream& out,
                                const deque<leaf_node *> expr) const;

  private:
};

class throw_expectation_node : public expectation_node
{
  public:
    throw_expectation_node (const string &f, int l)
      : expectation_node(f, l) { }
    virtual ~throw_expectation_node() { }

    virtual void print_with_prefix (const string &prefix,
				const UnciLexer *lexer = nullptr) const;
    virtual void generate (ostream& out,
                                const deque<leaf_node *> expr) const;

  private:
};

class about_expectation_node : public expectation_node
{
  public:
    about_expectation_node (const string &f, int l, 
                        deque<leaf_node *> v, deque<leaf_node *> t)
      : expectation_node(f, l), value(v), tolerance(t) { }
    virtual ~about_expectation_node();

    virtual void print_with_prefix (const string &prefix,
				const UnciLexer *lexer = nullptr) const;
    virtual void generate (ostream& out,
                                const deque<leaf_node *> expr) const;

  private:
    deque<leaf_node *> value;
    deque<leaf_node *> tolerance;
};

class array_expectation_node : public expectation_node
{
  public:
    array_expectation_node (const string &f, int l,
                        deque<leaf_node *> v)
      : expectation_node(f, l), value(v) { }
    virtual ~array_expectation_node();

    virtual void print_with_prefix (const string &prefix,
				const UnciLexer *lexer = nullptr) const;
    virtual void generate (ostream& out,
                                const deque<leaf_node *> expr) const;

  private:
    deque<leaf_node *> value;
};

class trace_expectation_node : public expectation_node
{
  public:
    trace_expectation_node (const string &f, int l)
      : expectation_node(f, l) { }
    virtual ~trace_expectation_node() { }

    virtual void print_with_prefix (const string &prefix,
				const UnciLexer *lexer = nullptr) const;
    virtual void generate (ostream& out,
                                const deque<leaf_node *> expr) const;

  private:
};

class compop_expectation_node : public expectation_node
{
  public:
    compop_expectation_node (const string &f, int l,
                                leaf_node * c, deque<leaf_node *> v)
      : expectation_node(f, l), comp(c), value(v) { }
    virtual ~compop_expectation_node();

    virtual void print_with_prefix (const string &prefix,
				const UnciLexer *lexer = nullptr) const;
    virtual bool validate_and_transform ();
    virtual stmt_node * validation_stmt (const deque<leaf_node *>& expression) const;
    virtual void generate (ostream& out,
                                const deque<leaf_node *> expr) const;

  private:
    leaf_node * comp;
    deque<leaf_node *> value;
};

class fixture_node : public ast_node
{
  public:
    fixture_node (const string &f, int l, const deque<stmt_node *>& d,
                               named_block_node* s, named_block_node* t)
      : ast_node(f, l),
        decls(d), setup_init(nullptr), setup(s), teardown(t) { }
    virtual ~fixture_node();

    virtual void print_with_prefix (const string &prefix,
				const UnciLexer *lexer = nullptr) const;
    virtual bool validate_and_transform ();
    virtual void generate (ostream& out) const;

  private:
    deque<stmt_node *> decls;
    named_block_node * setup_init;
    named_block_node * setup;
    named_block_node * teardown;
};

class named_block_node : public ast_node
{
  public:
    named_block_node (const string &f, int l, block_node * b)
      : ast_node(f, l), block(b) { }
    virtual ~named_block_node();

    virtual void print_with_prefix (const string &prefix,
			        const UnciLexer *lexer = nullptr) const;
    virtual bool validate_and_transform ();
    virtual void generate (ostream& out) const;

    friend class fixture_node;

  private:
    block_node * block;
};

class test_node : public ast_node
{
  public:
    test_node (const string &f, int l, string n, block_node* b)
      : ast_node(f, l), name(n), block(b) { }
    virtual ~test_node();

    virtual void print_with_prefix (const string &prefix,
				const UnciLexer *lexer = nullptr) const;
    virtual bool validate_and_transform ();
    virtual void generate (ostream& out) const;

    inline string get_name() { return name; }

  private:
    string name;
    block_node * block;
};


#endif
