#include "uncilexer.h"
#include <cstdio>
#include <cstdlib>
#include <sstream>
using namespace std;

#include "tokens.h"

UnciLexer::UnciLexer()
  : yyFlexLexer(), filename("<stdin>"), lineno(1)
{
  initReserved();
  add_symbol("");  // invalid symbol at symbol ID 0

  token = yylex();
  tokentext = string(YYText());
}

UnciLexer::UnciLexer(istream* is, string fname)
  : yyFlexLexer(is, &cout), filename(fname), lineno(1)
{
  initReserved();
  add_symbol("");  // invalid symbol at symbol ID 0

  token = yylex();
  tokentext = string(YYText());
}

void UnciLexer::hash_locate(const char *hash)
{
  char *end;
  lineno = strtol(hash+2, &end, 10);
  // end points to space before filename; skip space and doublequote
  filename = string(end + 2);
  filename.erase (filename.end()-2, filename.end()); // remove " and \n
}

int UnciLexer::consume()
{
  //print_token_table_line(stderr);
  prev_token = token;
  prev_tokentext = tokentext;
  prev_ws = ws;
  prev_semval = semval;
  prev_lineno = lineno;
  prev_filename = filename;

  // not necessary, but helps catch errors
  token = 0;
  tokentext = "";
  ws = "";
  semval = 0;

  token = yylex();
  tokentext = string(YYText());
  //fprintf (stderr, "Old %d new %d\n", prev_token, token);
  return prev_token;
}

void UnciLexer::print_token_table_line(FILE *OUT)
{
  fprintf (OUT, "%9s ", token_name(consumed_token()).c_str());
  if (consumed_token() == STMTKEY || consumed_token() >= IDENT) // asgnop, ident, literals
    fprintf (OUT, "%-6ld", consumed_value());
  else 
    fprintf (OUT, "      ");
  fprintf (OUT, " %-20s %3d <%s>\n", consumed_tokentext().c_str(), consumed_lineno(), consumed_ws().c_str());
}

string UnciLexer::token_name(int tok) const
{
  if (tok < 128)
  {
    ostringstream charstr;
    charstr << '\'' << (char) tok << '\'';
    return charstr.str();
  }
  map<int, string>::const_iterator loc = token_names.find(tok);
  if (loc == token_names.end())
  {
    ostringstream numstr;
    numstr << tok;
    return numstr.str();
  }
  else
    return loc->second;
}



int UnciLexer::find_or_bust(const map<string, int> &dict, const string& s)
{
  map<string, int>::const_iterator loc = dict.find(s);
  if (loc == dict.end())
    return -1;
  else
    return loc->second;
}

int UnciLexer::symbol(const string& s) const
{
  return find_or_bust(symbol_table, s);
}

int UnciLexer::get_string_id(const string& s) const
{
  return find_or_bust(string_table, s);
}

int UnciLexer::add_to_table(map<string, int> &dict, vector<string> &table,
                const string& s)
{
  int id = find_or_bust(dict, s);
  if (id == -1)
  {
    id = table.size();
    table.push_back(s);
    dict[s] = id;
  }
  return id;
}

int UnciLexer::add_symbol(const string& s)
{
  map<string,int>::const_iterator loc = reserved_words.find(s);
  if (loc != reserved_words.end())
    return loc->second;

  semval = add_to_table(symbol_table, symbol_table_by_int, s);

  return 501; //IDENT;
}

void UnciLexer::add_string(const string& s)
{
  semval = add_to_table(string_table, string_table_by_int, s);
}

int UnciLexer::deback_char (char *charlit)
{
  if (charlit[0] != '\\')
  {
    return charlit[0];
  }
  else
  {
    switch (charlit[1])
    {
      case 'a': return '\a';
      case 'b': return '\b';
      case 'f': return '\f';
      case 'n': return '\n';
      case 't': return '\t';
      case 'r': return '\r';
      case 'v': return '\v';
      case '?': return '\?';
      case '0': return 0;
      case '\\': return '\\';
      case '\'': return '\'';
      case '\"': return '\"';
      default: return -1;
    }
  }
}
