#include "unciparser.h"

#include <iostream>
#include <cstdlib>
using namespace std;


void UnciParser::syn_fail(const string &expected)
{
  cerr << "unci: ";
  if (expected == "")
    cerr << "Found unexpected ";
  else
    cerr << "Expected " << expected << ", found ";
  cerr << lexer.token_name(lexer.peek()) 
                        << " at " << lexer.peek_filename() 
                        << ":" << lexer.peek_lineno() << endl;
  //cerr << "Aborting parse" << endl;
  exit(1);
}

test_suite * UnciParser::parse()
{
  string filename = lexer.peek_filename();
  int lineno = lexer.peek_lineno();
  deque<ast_node *> prologue = parse_prologue();

  must_consume(TEST, "test suite declaration at top of file");

  must_consume(SUITE, "test suite declaration at top of file");

  auto suite_ident = must_match(IDENT,
                        "test suite declaration at top of file");

  must_consume('{', "'{' to begin test suite");

  auto the_fixture = parse_the_fixture_opt();
  auto the_tests = parse_the_tests_opt();

  if (the_fixture == nullptr and the_tests.empty()
      and lexer.peek() != '}')
  {
    syn_fail ("'fixture:' (or 'tests:') at start of test suite");
  }
  else if (the_tests.empty() and lexer.peek() != '}')
  {
    syn_fail ("'tests:' before beginning of test cases");
  }


  must_consume('}', "'}' to end test suite");

  // ; is optional
  if (lexer.peek() == ';')
    lexer.consume();
  
  return new test_suite(filename, lineno, prologue, suite_ident,
                        the_fixture, the_tests);
}

deque<ast_node *> UnciParser::parse_prologue()
{
  deque<ast_node *> decls;
  ast_node * cur;
  while (true)
  {
    if (lexer.peek() == USING)
      cur = parse_using_stmt();
    else if (lexer.peek() == HASHLINE)
      cur = match(HASHLINE);
    else
      break;
    decls.push_back(cur);
  }
  return decls;
}

simple_stmt_node * UnciParser::parse_using_stmt()
{
  string filename = lexer.peek_filename();
  int lineno = lexer.peek_lineno();

  leaf_node * cur = match(USING);
  if (cur == nullptr)
    return nullptr;

  deque<leaf_node *> elements = parse_junk();
  elements.push_front(cur);

  cur = must_match(';', "semicolon at end of 'using' statement");
  elements.push_back(cur);

  return new simple_stmt_node(filename, lineno, elements);
}

fixture_node * UnciParser::parse_the_fixture_opt()
{
  string filename = lexer.peek_filename();
  int lineno = lexer.peek_lineno();

  auto cur = match(FIXTURE);
  if (cur == nullptr)
    return nullptr;
  delete cur;

  cur = must_match(':', "':' after keyword 'fixture'");
  delete cur;

  deque<stmt_node *> stmts;
  while (true)
  {
    auto cur = parse_simple_stmt_opt();
    if (cur == nullptr)
      break;
    stmts.push_back(cur);
  }

  auto setup = parse_named_block_opt(SETUP);
  auto teardown = parse_named_block_opt(TEARDOWN);

  return new fixture_node(filename, lineno, stmts, setup, teardown);
}

deque<test_node *> UnciParser::parse_the_tests_opt() 
{
  deque<test_node *> tests;
  auto cur = match(TESTS);
  if (cur == nullptr)
    return tests;
  delete cur;
  cur = must_match(':', "':' after keyword 'tests'");
  delete cur;

  while (true)
  {
    auto test = parse_test_opt();
    if (test == nullptr)
      break;
    tests.push_back(test);
  }

  // only correct FOLLOW is right curly; give useful error
  if (lexer.peek() != '}')
  {
    if (lexer.peek() == EOFT)
      syn_fail ("'}' to end test suite");
    else if (tests.empty())
      syn_fail ("all content under 'tests:' to be inside 'test' blocks");
    else
      syn_fail ("test suite to end after 'test' blocks");
  }

  return tests;
}

test_node * UnciParser::parse_test_opt()
{
  string filename = lexer.peek_filename();
  int lineno = lexer.peek_lineno();

  auto cur = match(TEST);
  if (cur == nullptr)
    return nullptr;
  delete cur;

  leaf_node * name_node = must_match(IDENT,
                "identifier after keyword 'test', to name the test case");

  block_node * block = parse_block();
  if (name_node == nullptr)
    syn_fail ("Internal error: no block");

  return new test_node(filename, lineno, name_node->get_text(), block);
}


named_block_node * UnciParser::parse_named_block_opt(token_t head)
{
  string filename = lexer.peek_filename();
  int lineno = lexer.peek_lineno();

  auto name = match(head);
  if (name == nullptr)
    return nullptr;

  auto block = parse_block();
  if (block == nullptr)
    syn_fail ("Internal error: no block");

  return new named_block_node(filename, lineno, block);
}

block_node * UnciParser::parse_block()
{
  string filename = lexer.peek_filename();
  int lineno = lexer.peek_lineno();

  auto open = must_match('{', "'{' to begin block");

  deque<stmt_node *> stmts;
  while (true)
  {
    stmt_node * cur = parse_stmt_opt();
    if (cur == nullptr)
      break;
    stmts.push_back(cur);
  }

  auto close = must_match('}', "'}' to close block, or more statements");

  return new block_node(filename, lineno, open, stmts, close);
}

stmt_node * UnciParser::parse_stmt(bool opt)
{
  string filename = lexer.peek_filename();
  int lineno = lexer.peek_lineno();

  switch (lexer.peek())
  {
    case CHECK: return parse_check_stmt();
    case ';':   return new empty_stmt_node(filename, lineno,
                  must_match(';', "Internal error: empty stmt"));
    case '{':   return parse_block();
    case STMTKEY:  
                return parse_control_stmt();

    case STRLIT:
    case NUMLIT:
    case CHARLIT:
    case IDENT:
    case OPPUNC:

    case TRUE:
    case FALSE:
    case THROW:
    case EXPECT:
    case ABOUT:
    case ARRAY:
    case TRACE:
    case FIXTURE:
    case SUITE:

    case '(':
                return parse_simple_stmt();

    default:
                if (opt)
                  return nullptr;
                else
                  syn_fail ("beginning of statement");
  }
  return nullptr; //never reached
}

simple_stmt_node * UnciParser::parse_simple_stmt(bool opt)
{
  string filename = lexer.peek_filename();
  int lineno = lexer.peek_lineno();

  switch (lexer.peek())
  {
    case STRLIT:
    case NUMLIT:
    case CHARLIT:
    case IDENT:
    case OPPUNC:

    case TRUE:
    case FALSE:
    case THROW:
    case EXPECT:
    case ABOUT:
    case ARRAY:
    case TRACE:
    case FIXTURE:
    case SUITE:

    case '(':
                // there is a simple stmt to parse
                break;

    default:
                if (opt)
                  return nullptr;
                else
                  syn_fail ("beginning of simple statement");
  }

  auto elements = parse_junk(false);

  auto cur = must_match(';', "';' to end statement");
  elements.push_back(cur);

  return new simple_stmt_node(filename, lineno, elements);
}

check_stmt_node * UnciParser::parse_check_stmt()
{
  string filename = lexer.peek_filename();
  int lineno = lexer.peek_lineno();

  auto check = must_match(CHECK, "check/expect statement");
  string ws = check->get_whitespace();
  delete check;

  auto left = must_match('(', "'(' to begin expression to be checked");
  auto junk = parse_junk();
  auto right = must_match(')', "')' to conclude expression to be checked");
  junk.push_front(left);
  junk.push_back(right);


  must_consume(EXPECT, "'expect' to begin expectation");
  // ws? XXX

  auto expectation = parse_expectation();

  must_consume(';', "';' after check/expect statement");
  // ws? XXX

  return new check_stmt_node(filename, lineno, ws, junk, expectation);
}

// stmt_key_tail
control_stmt_node * UnciParser::parse_control_stmt()
{
  string filename = lexer.peek_filename();
  int lineno = lexer.peek_lineno();

  auto keyword = must_match(STMTKEY,
                        "control keyword to begin control statement");

  deque<leaf_node *> condition;
  auto cur = match('(');
  if (cur != nullptr)
  {
    condition.push_back(cur);
    while (true)
    {
      auto junk = parse_junk_opt();
      if (not junk.empty())
      {
        for (auto elt : junk)
          condition.push_back(elt);
      }
      else if (lexer.peek() == ';')
      {
        condition.push_back(must_match(';', "Internal error: ;"));
      }
      else break;
    }
    cur = must_match(')', "')' to close the control statement condition");
    condition.push_back(cur);
  }

  auto block = parse_stmt();

  return new control_stmt_node(filename, lineno, keyword, condition, block);
}

expectation_node * UnciParser::parse_expectation()
{
  string filename = lexer.peek_filename();
  int lineno = lexer.peek_lineno();

  switch (lexer.peek())
  {
    case TRUE:
                must_consume(TRUE, "Internal error: exp");
                return new true_expectation_node(filename, lineno);
    case FALSE:
                must_consume(FALSE, "Internal error: exp");
                return new false_expectation_node(filename, lineno);
    case THROW:
                must_consume(THROW, "Internal error: exp");
                return new throw_expectation_node(filename, lineno);
    case ABOUT:
                return parse_about_expectation();
    case ARRAY:
                return parse_array_expectation();
    case TRACE:
                must_consume(TRACE, "Internal error: exp");
                return new trace_expectation_node(filename, lineno);
    case COMPOP:
                return parse_compop_expectation();
    default:
      syn_fail("expected results after keyword 'expect' (did you forget an == before an expected value?)");
      return nullptr;
  }

  //return new expectation_node (filename, lineno); //XXX
}

about_expectation_node * UnciParser::parse_about_expectation()
{
  string filename = lexer.peek_filename();
  int lineno = lexer.peek_lineno();

  must_consume (ABOUT, "'about' to begin expectation");
  auto value = parse_junk();
  
  must_consume(PLUSMINUS,
      "'+-' after expected approximate value, to set permitted tolerance");
  auto tolerance = parse_junk();
  if (tolerance.empty())
    syn_fail("expression after '+-' to indicate permitted tolerance (i.e. plus or minus how much?)");

  return new about_expectation_node(filename, lineno, value, tolerance);
}

array_expectation_node * UnciParser::parse_array_expectation()
{
  string filename = lexer.peek_filename();
  int lineno = lexer.peek_lineno();

  must_consume(ARRAY, "'array' to begin expectation");
  auto value = parse_junk();
  return new array_expectation_node(filename, lineno, value);
}

compop_expectation_node * UnciParser::parse_compop_expectation()
{
  string filename = lexer.peek_filename();
  int lineno = lexer.peek_lineno();

  auto comp = must_match(COMPOP, 
        "comparison operator (such as ==) to begin expectation");
  auto value = parse_junk();
  return new compop_expectation_node(filename, lineno, comp, value);
}


deque<leaf_node *> UnciParser::parse_junk(bool stmt_word_ok)
{
  deque<leaf_node *> elements;

  switch (lexer.peek())
  {
    case CHECK:
    case SETUP:
    case TEARDOWN:
    case TEST:
    case TESTS:
                if (!stmt_word_ok)
                  syn_fail("C++ expression (not unci reserved word)");
                /* else fall through */
    case STRLIT:
    case NUMLIT:
    case CHARLIT:
    case IDENT:
    case OPPUNC:

    case TRUE:
    case FALSE:
    case THROW:
    case EXPECT:
    case ABOUT:
    case ARRAY:
    case TRACE:
    case FIXTURE:
    case SUITE:
                elements.push_back(parse_leaf());
                return parse_junk_tail(elements);

    case '(':
                for (auto elt : parse_junk_matched(parse_leaf()))
                  elements.push_back(elt);
                return parse_junk_tail(elements);

    default:
                syn_fail("C++ expression");
                
  }
  return elements;
}

deque<leaf_node *> UnciParser::parse_junk_matched(leaf_node * left_bracket)
{
  int right_bracket_token;
  switch (left_bracket->get_token())
  {
    case '(': right_bracket_token = ')'; break;
    case '[': right_bracket_token = ']'; break;
    case '{': right_bracket_token = '}'; break;
    default: syn_fail("Internal error: unknown left bracket");
  }

  deque<leaf_node *> elements = parse_junk_opt();
  elements.push_front(left_bracket);

  leaf_node * right_bracket = match( right_bracket_token );
  if (right_bracket == nullptr)
  {
    syn_fail(string("Matching ") + (char)right_bracket_token);
  }
  elements.push_back(right_bracket);

  return elements;
}

// should be de-recursivised XXX
deque<leaf_node *> UnciParser::parse_junk_tail(deque<leaf_node *>& elements)
{
  switch (lexer.peek())
  {
    case CHECK:
    case SETUP:
    case TEARDOWN:
    case TESTS:

    case STRLIT:
    case NUMLIT:
    case CHARLIT:
    case IDENT:
    case OPPUNC:

    case TRUE:
    case FALSE:
    case THROW:
    case EXPECT:
    case ABOUT:
    case ARRAY:
    case TRACE:
    case FIXTURE:
    case TEST:
    case SUITE:

    case COMPOP:
    case ':':
                elements.push_back(parse_leaf());
                return parse_junk_tail(elements);

    case '(':
    case '[':
    case '{':
                for (auto elt : parse_junk_matched(parse_leaf()))
                  elements.push_back(elt);
                return parse_junk_tail(elements);

    default:
                // tail was empty, return elements so far
                return elements;
  }
}

deque<leaf_node *> UnciParser::parse_junk_opt()
{
  deque<leaf_node *> elements;
  switch (lexer.peek())
  {
    case CHECK:
    case SETUP:
    case TEARDOWN:
    case TESTS:

    case STRLIT:
    case NUMLIT:
    case CHARLIT:
    case IDENT:
    case OPPUNC:

    case TRUE:
    case FALSE:
    case THROW:
    case EXPECT:
    case ABOUT:
    case ARRAY:
    case TRACE:
    case FIXTURE:
    case TEST:
    case SUITE:

    case '(':
                return parse_junk();

    default:
                return elements;
  }
}


leaf_node * UnciParser::parse_leaf()
{
  lexer.consume();
  return new leaf_node(lexer.consumed_filename(), lexer.consumed_lineno(),
      lexer.consumed_token(), lexer.consumed_tokentext(), lexer.consumed_ws());
}

leaf_node * UnciParser::match(token_t tok)
{
  if (lexer.peek() == tok)
    return parse_leaf();
  else
    return nullptr;
}

leaf_node * UnciParser::must_match(token_t tok, string errmsg)
{
  if (lexer.peek() == tok)
    return parse_leaf();
  else
    syn_fail(errmsg);
  return nullptr; // never reached;
}

bool UnciParser::must_consume(token_t tok, string errmsg)
{
  if (lexer.peek() == tok)
  {
    lexer.consume();
    return true;
  }
  else
  {
    syn_fail(errmsg);
    return false; // never reached;
  }
}
